#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, annotations

from hashlib import sha256

from sklearn.metrics.pairwise import cosine_similarity
from os.path import basename, exists, expanduser
from requests import request, Response, HTTPError
import requests
from numpy import array, vstack, dot, arccos, pi
from pandas import DataFrame
import json
import tomli
import logging
from lief import parse, ELF, PE, MachO


re_conf = {
    "apikey": "l1br3",
    "host": "https://api.reveng.ai",
    "model": "binnet-0.2-x86"
}


logger = logging.getLogger("REAIT")


def reveng_req(r: request, end_point: str, data=None, ex_headers: dict = None, params=None,
               json_data: dict = None, timeout: int = 30) -> Response:
    """
    Constructs and sends a Request
    :param r: Method for the new Request
    :param end_point: Endpoint to add to the base URL
    :param ex_headers: Extended HTTP headers to add
    :param data: Dictionary, list of tuples, bytes, or file-like object to send in the body
    :param params: Dictionary, list of tuples or bytes to send in the query string for the query string
    :param json_data: A JSON serializable Python object to send in the body
    :param timeout: Number of seconds to stop waiting for a Response
    """
    url = f"{re_conf['host']}/{end_point}"
    headers = {"Authorization": re_conf["apikey"]}

    if ex_headers:
        headers.update(ex_headers)

    logger.debug("Making request %s:\n  - headers: %s\n  - data: %s\n  - json_data: %s\n  - params: %s",
                 url, headers, data, json_data, params)

    response: Response = r(url, headers=headers, json=json_data, data=data, params=params, timeout=timeout)

    logger.debug("Making response %s:\n  - headers: %s\n  - status_code: %d\n  - content: %s",
                 url, response.headers, response.status_code, response.text)

    return response


def re_hash_check(bin_id: str) -> bool:
    status = False
    res = reveng_req(requests.get, f"search?search=sha_256_hash:{bin_id}&state=All&user_owned=true")

    if res.status_code == 200:
        binaries_data = res.json()["binaries"]
        status = len(binaries_data) > 0
    elif res.status_code == 400:
        logger.warning("Bad Request: %s", res.text)
    else:
        logger.error("Internal Server Error.")

    res.raise_for_status()
    return status


# Bin_id is referred to as hash in this program - to maintain usage BID = id of a binary bin_id = hash
# Assumes a file has been passed, correct hash only
# Returns the BID of the binary_id (hash)
def re_bid_search(bin_id: str) -> int:
    res = reveng_req(requests.get, f"search?search=sha_256_hash:{bin_id}&state=All")

    bid = -1

    # Valid request
    if res.status_code == 200:
        # Check only one record is returned
        binaries_data = res.json()["binaries"]

        if len(binaries_data) > 1:
            logger.info("%d matches found for hash: %s.", len(binaries_data), bin_id)

            if len(binaries_data) > 1:
                options_dict = {}

                for idx, binary in enumerate(binaries_data):
                    logger.info("[%d] - ID: {}, Name: %s, Creation: %s, Model: %s, Owner: %s, Status: %s",
                                idx, binary["binary_id"], binary["binary_name"], binary["creation"],
                                binary["model_name"], binary["owner"], binary["status"])

                    options_dict[idx] = binary["binary_id"]

                user_input = input("[+] Please enter the option you want to use for this operation:")

                try:
                    option_number = int(user_input)

                    bid = options_dict.get(option_number, -1)

                    if bid == -1:
                        logger.warning("Invalid option.")
                except Exception:
                    bid = -1
                    logger.warning("Invalid option.")
            # Only 1 match found
            elif len(binaries_data) == 1:
                binary = binaries_data[0]
                bid = binary["binary_id"]
            else:
                logger.warning("No matches found for hash: %s.", bin_id)
        elif len(binaries_data) == 1:
            binary = binaries_data[0]
            bid = binary["binary_id"]

            logger.info("Only one record exists, selecting - ID: %d, Name: %s, "
                        "Creation: %s, Model: %s, Owner: %s, Status: %s",
                        bid, binary["binary_name"], binary["creation"],
                        binary["model_name"], binary["owner"], binary["status"])
        else:
            logger.warning("No matches found for hash: %s.", bin_id)
    elif res.status_code == 400:
        logger.warning("Bad Request: %s", res.text)
    else:
        logger.error("Internal Server Error.")

    res.raise_for_status()
    return bid


def RE_delete(fpath: str, binary_id: int = 0) -> Response:
    """
    Delete analysis results for Binary ID in command
    :param fpath: File path for binary to analyse
    :param binary_id: ID of binary
    """
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id == 0 else binary_id

    if bid == -1:
        raise HTTPError(f"No matches found for hash: {bin_id}")

    res = reveng_req(requests.delete, f"analyse/{bid}")

    if res.status_code == 200:
        logger.info("Securely deleted %s analysis.", bin_id)
    elif res.status_code == 404:
        logger.warning("Error analysis not found for %s.", bin_id)
    else:
        logger.error("Error deleting binary %s under. Server returned %d.", bin_id, res.status_code)

    res.raise_for_status()
    return res


def RE_analyse(fpath: str, model_name: str = None, isa_options: str = None, platform_options: str = None,
               file_options: str = None, dynamic_execution: bool = False, command_line_args: str = None,
               scope: str = None, tags: list = None, priority: int = 0,
               duplicate: bool = False, symbols: dict = None) -> Response:
    """
    Start analysis job for binary file
    :param fpath: File path for binary to analyse
    :param model_name: Binary model name
    :param isa_options: Executable ISA
    :param file_options: File options
    :param platform_options: OS platform
    :param dynamic_execution: Enable dynamic execution in sandbox during analysis
    :param command_line_args: Command line arguments to pass when running binary sample in the sandbox
    :param scope: Analysis visibility
    :param tags: Assign tags to an analysis
    :param priority: Priority to processing queue
    :param duplicate: Duplicate an existing binary
    :param symbols: List of functions
    """
    bin_id = re_binary_id(fpath)
    result = re_hash_check(bin_id)

    if result and duplicate is False:
        logger.error("Error, duplicate analysis for %s. To upload again, use the --duplicate flag.",
                     bin_id)
        raise HTTPError(f"Duplicate analysis for hash: {bin_id}")

    filename = basename(fpath)

    params = {"file_name": filename, "sha_256_hash": bin_id}

    for p_name in ("model_name", "isa_options", "platform_options", "file_options",
                   "dynamic_execution", "command_line_args", "scope", "tags", "priority", "symbols"):
        p_value = locals()[p_name]

        if p_value:
            params[p_name] = p_value

    res = reveng_req(requests.post, f"analyse", json_data=params)

    if res.status_code == 200:
        logger.info("Successfully submitted binary for analysis. %s - %s", fpath, re_binary_id(fpath))
    elif res.status_code == 400:
        response = res.json()

        if "error" in response.keys():
            logger.warning("Error analysing %s - %s", fpath, response["error"])

    res.raise_for_status()
    return res


def RE_upload(fpath: str) -> Response | bool:
    """
    Upload binary to Server
    :param fpath: File path for binary to analyse
    """
    bin_id = re_binary_id(fpath)
    result = re_hash_check(bin_id)

    if result:
        logger.info("File %s - %s already exists. Skipping upload...", basename(fpath), re_binary_id(fpath))
        return True

    res = reveng_req(requests.post, f"upload", data=open(fpath, "rb").read())

    if res.status_code == 200:
        logger.info("Successfully uploaded binary to your account. %s - %s", fpath, re_binary_id(fpath))
    elif res.status_code == 400:
        response = res.json()

        if "error" in response.keys():
            logger.warning("Error uploading %s - %s", fpath, response["error"])
    elif res.status_code == 413:
        logger.warning("File too large. Please upload files under 100MB.")
    elif res.status_code == 500:
        logger.error("Internal Server Error. Please contact support. Skipping upload...")

    res.raise_for_status()
    return res


def RE_embeddings(fpath: str, binary_id: int = 0) -> Response:
    """
    Fetch symbol embeddings
    :param fpath: File path for binary to analyse
    :param binary_id: ID of binary
    """
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id == 0 else binary_id

    if bid == -1:
        raise HTTPError(f"No matches found for hash: {bin_id}")

    res = reveng_req(requests.get, f"embeddings/{bid}")

    if res.status_code == 400:
        logger.warning("Analysis for %s still in progress. Please check the logs (-l) and try again later.",
                       bin_id)

    res.raise_for_status()
    return res


def RE_signature(fpath: str, binary_id: int = 0) -> Response:
    """
    Fetch binary BinNet signature
    :param fpath: File path for binary to analyse
    :param binary_id: ID of binary
    """
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id == 0 else binary_id

    if bid == -1:
        raise HTTPError(f"No matches found for hash: {bin_id}")

    res = reveng_req(requests.get, f"signature/{bid}")

    if res.status_code == 425:
        logger.warning("Analysis for %s still in progress. Please check the logs (-l) and try again later.",
                       bin_id)

    res.raise_for_status()
    return res


def RE_embedding(fpath: str, start_vaddr: int, end_vaddr: int = None, base_vaddr: int = None,
                 model: str = None) -> Response:
    """
    Fetch embedding for custom symbol range
    :param fpath: File path for binary to analyse
    :param start_vaddr: Start virtual address of the function to extract embeddings
    :param end_vaddr: End virtual address of the function to extract embeddings
    :param base_vaddr: Base address of the binary
    :param model: Binary model name
    """
    params = {}

    if end_vaddr:
        params["end_vaddr"] = end_vaddr
    if base_vaddr:
        params["base_vaddr"] = base_vaddr
    if model:
        params["models"] = model

    bin_id = re_binary_id(fpath)

    res = reveng_req(requests.get, f"embedding/{bin_id}/{start_vaddr}", params=params)

    if res.status_code == 425:
        logger.warning("Analysis for %s still in progress. Please check the logs (-l) and try again later.",
                       bin_id)

    res.raise_for_status()
    return res


def RE_logs(fpath: str, binary_id: int = 0, console: bool = True) -> Response:
    """
    Get the logs for an analysis associated to Binary ID in command
    :param fpath: File path for binary to analyse
    :param binary_id: ID of binary
    :param console: Show response in console
    """
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id == 0 else binary_id

    if bid == -1:
        raise HTTPError(f"No matches found for hash: {bin_id}")

    res = reveng_req(requests.get, f"logs/{bid}")

    if res.status_code == 200 and console:
        logger.info("Logs found for %s:\n%s", bin_id, res.text)
    elif res.status_code == 404:
        logger.warning("Error, logs not found for %s.", bin_id)

    res.raise_for_status()
    return res


def RE_cves(fpath: str, binary_id: int = 0) -> Response:
    """
    Check for known CVEs in Binary
    :param fpath: File path for binary to analyse
    :param binary_id: ID of binary
    """
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id == 0 else binary_id

    if bid == -1:
        raise HTTPError(f"No matches found for hash: {bin_id}")

    res = reveng_req(requests.get, f"cves/{bid}")

    if res.status_code == 200:
        cves = json.loads(res.text)
        logger.info("Checking for known CVEs embedded inside %s", fpath)

        if len(cves) == 0:
            logger.info("0 CVEs found.")
        else:
            logger.warning("Warning CVEs found!\n%s", res.text)
    elif res.status_code == 404:
        logger.warning("Error, binary analysis not found for %s.", bin_id)

    res.raise_for_status()
    return res


def RE_status(fpath: str, binary_id: int = 0) -> Response:
    """
    Get the status of an ongoing binary analysis
    :param fpath: File path for binary to analyse
    :param binary_id: ID of binary
    """
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id == 0 else binary_id

    if bid == -1:
        raise HTTPError(f"No matches found for hash: {bin_id}")

    res = reveng_req(requests.get, f"analyse/status/{bid}")

    if res.status_code == 400:
        logger.warning(" Error, status not found for %s.", bin_id)

    res.raise_for_status()
    return res


def RE_compute_distance(embedding: list, embeddings: list, nns: int = 5) -> list:
    """
    Compute the cosine distance between source embedding and embedding from binary
    :param embedding: Embedding vector as python list
    :param embeddings: Symbol embeddings
    :param nns: Number of nearest neighbors
    """
    df = DataFrame(data=embeddings)
    np_embedding = array(embedding).reshape(1, -1)
    source_embeddings = vstack(df["embedding"].values)
    closest = cosine_similarity(source_embeddings, np_embedding).squeeze().argsort()[::-1][:nns]
    distances = cosine_similarity(source_embeddings[closest], np_embedding)

    # match closest embeddings with similarity
    closest_df = df.iloc[closest]

    # create json similarity object
    similarities = list(zip(distances, closest_df.index.tolist()))
    json_sims = [{"similaritiy": float(d[0]), "vaddr": int(df.iloc[v]["vaddr"]), "name": str(df.iloc[v]["name"]),
                  "size": int(df.iloc[v]["size"])} for d, v in similarities]
    return json_sims


def RE_nearest_symbols(embedding: list, model_name: str, nns: int = 5,
                       collections: list = None, ignore_hashes: list = None,
                       distance: float = 0.0, debug_enabled: bool = False) -> Response:
    """
    Get function name suggestions for an embedding
    :param embedding: Embedding vector as python list
    :param model_name: Binary model name
    :param nns: Number of nearest neighbors
    :param collections: List of collections RevEng.AI collection names to search through
    :param ignore_hashes: List[str] SHA-256 hash of binary file to ignore symbols from (usually the current binary)
    :param distance: How close we want the ANN search to filter for
    :param debug_enabled: ANN Symbol Search, only perform ANN on debug symbols if set
    """
    params = {"nns": nns, "model_name": model_name, "debug_enabled": debug_enabled}

    if collections and len(collections) > 0:
        # api param is collection, not collections
        params["collection"] = "|".join(collections)

    if ignore_hashes and len(ignore_hashes) > 0:
        params["ignore_hashes"] = ignore_hashes

    if distance > 0.0:
        params["distance"] = distance

    res = reveng_req(requests.post, "ann/symbol", data=json.dumps(embedding), params=params)

    res.raise_for_status()
    return res


def RE_nearest_binaries(embedding: list, model_name: str, nns: int = 5,
                        collections: list = None, ignore_hashes: list = None) -> Response:
    """
    Get executable suggestions for a binary embedding
    :param embedding: Embedding vector as python list
    :param model_name: Binary model name
    :param nns: Number of nearest neighbors
    :param collections: List of collections RevEng.AI collection names to search through
    :param ignore_hashes: List[str] SHA-256 hash of binary files to ignore symbols from (usually the current binary)
    """
    params = {"nns": nns, "model_name": model_name}

    if collections and len(collections) > 0:
        # api param is collection, not collections
        params["collection"] = "|".join(collections)

    if ignore_hashes and len(ignore_hashes) > 0:
        params["ignore_hashes"] = ignore_hashes

    res = reveng_req(requests.post, "ann/binary", data=json.dumps(embedding), params=params)

    res.raise_for_status()
    return res


def RE_SBOM(fpath: str, binary_id: int = 0) -> Response:
    """
    Get Software Bill Of Materials for binary
    :param fpath: File path for binary to analyse
    :param binary_id: ID of binary
    """
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id == 0 else binary_id

    if bid == -1:
        raise HTTPError(f"No matches found for hash: {bin_id}")

    res = reveng_req(requests.get, f"sboms/{bid}")

    logger.info("SBOM for %s:\n%s", fpath, res.text)

    res.raise_for_status()
    return res


def RE_functions_rename(function_id: int, new_name: str) -> Response:
    """
    Send the new name of a function to C2
    :param function_id: ID of a function
    :param new_name: New function name
    """
    res = reveng_req(requests.post, f"functions/rename/{function_id}", json_data={"new_name": new_name})

    if res.status_code == 200:
        logger.info("FunctionId %d has been renamed with '%s'.", function_id, new_name)
    else:
        logger.warning("Error, cannot rename FunctionId %d. %s", function_id, res.text)

    res.raise_for_status()
    return res


def re_binary_id(fpath: str) -> str:
    """
    Take the SHA-256 hash of binary file
    :param fpath: File path for binary to analyse
    """
    if not fpath or not exists(fpath):
        return "undefined"

    hf = sha256()

    with open(fpath, "rb") as f:
        c = f.read()
        hf.update(c)

    return hf.hexdigest()


def _binary_isa(lief_hdlr, exec_type: str) -> str:
    """
    Get ISA format
    """
    if exec_type == "elf":
        machine_type = lief_hdlr.header.machine_type

        if machine_type == ELF.ARCH.i386:
            return "x86"
        elif machine_type == ELF.ARCH.x86_64:
            return "x86_64"
    elif exec_type == "pe":
        machine_type = lief_hdlr.header.machine

        if machine_type == PE.MACHINE_TYPES.I386:
            return "x86"
        elif machine_type == PE.MACHINE_TYPES.AMD64:
            return "x86_64"
    elif exec_type == "macho":
        machine_type = lief_hdlr.header.cpu_type

        if machine_type == MachO.CPU_TYPES.x86:
            return "x86"
        elif machine_type == MachO.CPU_TYPES.x86_64:
            return "x86_64"

    logger.error("Error, failed to determine or unsupported ISA for exec_type: %s.", exec_type)
    raise RuntimeError(f"Error, failed to determine or unsupported ISA for exec_type:{exec_type}.")


def _binary_format(lief_hdlr) -> str:
    """
    Get executable file format
    """
    if lief_hdlr.format == lief_hdlr.format.PE:
        return "pe"
    if lief_hdlr.format == lief_hdlr.format.ELF:
        return "elf"
    if lief_hdlr.format == lief_hdlr.format.MACHO:
        return "macho"

    logger.error("Error, could not determine binary format: %s.", lief_hdlr.format)
    raise RuntimeError("Error, could not determine binary format.")


def file_type(fpath: str) -> tuple[str, str]:
    """
    Determine ISA for binary
    :param fpath: File path for binary to analyse
    """
    binary = parse(fpath)

    # handle PE and ELF files
    file_format = _binary_format(binary)
    isa = _binary_isa(binary, file_format)

    return file_format, isa


def parse_config() -> None:
    """
    Parse ~/.reait.toml config file
    """
    if exists(expanduser("~/.reait.toml")):
        with open(expanduser("~/.reait.toml"), "r") as file:
            config = tomli.loads(file.read())

            for key in ("apikey", "host", "model"):
                if key in config:
                    re_conf[key] = config[key]


def angular_distance(x, y) -> float:
    """
    Compute angular distance between two embedding vectors
    Normalised euclidian distance
    """
    cos = dot(x, y) / ((dot(x, x) * dot(y, y)) ** 0.5)
    return 1.0 - arccos(cos) / pi



#!/usr/bin/env python
from __future__ import print_function

import logging

from rich import print_json
from rich.progress import track
from rich.console import Console
from rich.table import Table
import os
import argparse
import json
from os.path import isfile
from sys import exit, stdout, stderr
from reait import api, __version__
from scipy.spatial import distance
from glob import iglob
import numpy as np
from concurrent.futures import ThreadPoolExecutor, as_completed
from multiprocessing import cpu_count

rerr = Console(file=stderr)
rout = Console(file=stdout)


def version():
    """
        Display program version
    """
    rout.print(f"""[bold blue] ::::::::    ::::::::
::  ::::    :::  :::
::::::::::::::::::::
:::::   :::   ::::::
   ::::::::::::::   
   .::  :::  ::::   
::::::  :::  :::::::
::  :::::::::::  :::
::  :::::  ::::  :::
::::::::    :::::::: [/bold blue]
  [bold red]reait[/bold red] [bold bright_green]v{__version__}[/bold bright_green]
""")
    rout.print("[yellow]Config:[/yellow]")
    print_json(data=api.re_conf)


def verify_binary(fpath_fmt: str):
    fmt = None
    fpath = fpath_fmt

    # if ':' in fpath_fmt:
    #    fpath, fmt = fpath_fmt.split(':')

    if not os.path.isfile(fpath):
        raise RuntimeError(f"File path {fpath} is not a file")

    # if getsize(fpath) > 1024 * 1024 * 10:
    #    raise RuntimeError("Refusing to analyse file over 10MB. Please use a RevEng.AI SRE integration")

    if not fmt:
        exec_format, exec_isa = api.file_type(fpath)
    else:
        if '-' not in fmt:
            raise RuntimeError(
                'Binary type must follow format {EXEC_FORMAT}-{ISA}. Use EXEC_FORMAT raw for memory dumps e.g. raw-x86')

        exec_format, exec_isa = fmt.split('-')

    return fpath, exec_format, exec_isa


def match(fpath: str, model_name: str, embeddings: list, confidence: float = 0.95, deviation: float = 0.1):
    """
    Match embeddings in fpath from a list of embeddings
    """
    print(f"Matching symbols from {fpath} with confidence {confidence}")
    sink_embed_mat = np.vstack(list(map(lambda x: x['embedding'], embeddings)))
    b_embeds = api.RE_embeddings(fpath).json()
    source_embed_mat = np.vstack(list(map(lambda x: x['embedding'], b_embeds)))
    # angular distance over cosine 
    # closest = 1.0 - distance.cdist(source_embed_mat, sink_embed_mat, 'cosine')
    closest = distance.cdist(source_embed_mat, sink_embed_mat, api.angular_distance)
    # rescale to separate high end of (-1, 1.0)
    # closest = rescale_sim(closest)
    i, j = closest.shape

    for _i in track(range(i), description='Matching Symbols...'):
        row = closest[_i, :]
        match_index, second_match = row.argsort()[::-1][:2]
        source_index = _i
        sink_index = match_index
        source_symb = b_embeds[_i]
        sink_symb = embeddings[sink_index]
        m_confidence = row[match_index]
        s_confidence = row[second_match]

        if row[match_index] >= confidence:
            rout.print(
                f"[bold green]Found match![/bold green][yellow]\tConfidence: {m_confidence:.05f}[/yellow]\t[blue]{source_symb['name']}:{source_symb['vaddr']}[/blue]\t->\t[blue]{sink_symb['name']}:{sink_symb['vaddr']}")
        elif (m_confidence - s_confidence) > deviation:
            rout.print(
                f"[bold magenta]Possible match[/bold magenta][yellow]\tConfidence: {m_confidence:.05f}/{s_confidence:.05f}[/yellow]\t[blue]{source_symb['name']}:{source_symb['vaddr']}[/blue]\t->\t[blue]{sink_symb['name']}:{sink_symb['vaddr']}")
        else:
            rerr.print(
                f"[bold red]No match for[/bold red]\t[blue]{source_symb['name']}:{source_symb['vaddr']}\t{sink_symb['name']} - {m_confidence:0.05f}[/blue]")
            pass


def match_for_each(fpath: str, model_name: str, confidence: float = 0.95, collections=None):
    """
    Match embeddings in fpath from a list of embeddings
    """
    if collections is None:
        collections = []
    print(f"Matching symbols from {fpath} with confidence {confidence}")
    b_embeds = api.RE_embeddings(fpath).json()
    b_hash = api.re_binary_id(fpath)

    with ThreadPoolExecutor(max_workers=cpu_count()) as p:
        # print(f"Collections: {collections}")
        partial = lambda x: api.RE_nearest_symbols(x['embedding'], model_name, 1, collections=collections,
                                                   ignore_hashes=[b_hash]).json()
        res = {p.submit(partial, embed): embed for embed in b_embeds}

        for future in track(as_completed(res), description='Matching Symbols...'):
            # get result from future
            symbol = res[future]

            embedding = symbol['embedding']
            # do ANN call to match symbols, ignore functions from current file
            f_suggestions = api.RE_nearest_symbols(embedding, model_name, 1, collections=collections,
                                                   ignore_hashes=[api.re_binary_id(fpath)]).json()

            if len(f_suggestions) == 0:
                # no match
                rerr.print(f"\t[bold red]No match for[/bold red]\t[blue]{symbol['name']}:{symbol['vaddr']}[/blue]")
                continue

            matched = f_suggestions[0]
            if matched['distance'] >= confidence:
                rout.print(
                    f"\t[bold green]Found match![/bold green][yellow]\tConfidence: {matched['distance']:.05f}[/yellow]\t[blue]{symbol['name']}:{symbol['vaddr']}[/blue]\t->\t[blue]{matched['name']}:{matched['sha_256_hash']}")
                continue

            rerr.print(f"\t[bold red]No match for[/bold red]\t[blue]{symbol['name']}:{symbol['vaddr']}[/blue]")


def parse_collections(collections: str):
    """
        Return collections as list from CSV
    """
    if not collections:
        return None
    return collections.split(',')


def rescale_sim(x):
    """
        Too many values close to 0.999, 0.99999, 0.998, rescale so small values are very low, high values separated, map to hyperbolic space
    """
    return np.power(x, 5)


def binary_similarity(fpath: str, fpaths: list, model_name: str):
    """
    Compute binary similarity between source and list of binary files
    """
    console = Console()

    table = Table(title=f"Binary Similarity to {fpath}")
    table.add_column("Binary", justify="right", style="cyan", no_wrap=True)
    table.add_column("SHA3-256", style="magenta", no_wrap=True)
    table.add_column("Similarity", style="yellow", no_wrap=True)

    b_embed = api.RE_signature(fpath).json()

    b_sums = []
    for b in track(fpaths, description='Computing Binary Similarity...'):
        try:
            b_sum = api.RE_signature(b).json()
            b_sums.append(b_sum)
        except Exception as e:
            rerr.print(f"\n[red bold]{b} Not Analysed[/red bold] - [green bold]{api.re_binary_id(b)}[/green bold]")
            rerr.print(e)

    if len(b_sums) > 0:
        # closest = 1.0 - distance.cdist(np.expand_dims(b_embed, axis=0), np.vstack(b_sums), 'cosine')
        closest = distance.cdist(np.expand_dims(b_embed, axis=0), np.vstack(b_sums), api.angular_distance)

        for binary, similarity in zip(fpaths, closest.tolist()[0]):
            # table.add_row(os.path.basename(binary), api.re_binary_id(binary), f"{rescale_sim(similarity):.05f}")
            table.add_row(os.path.basename(binary), api.re_binary_id(binary), f"{similarity:.05f}")

    rout.print(table)


def main() -> None:
    """
    Tool entry
    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("-b", "--binary", default="",
                        help="Path of binary to analyse, use ./path:{exec_format} to specify executable format e.g. ./path:raw-x86_64")
    parser.add_argument("-B", "--binary-hash", default="", help="Hex-encoded SHA-256 hash of the binary to use")
    parser.add_argument("-D", "--dir", default="", help="Path of directory to recursively analyse")
    parser.add_argument("-a", "--analyse", action='store_true',
                        help="Perform a full analysis and generate embeddings for every symbol")
    parser.add_argument("--no-embeddings", action='store_true',
                        help="Only perform binary analysis. Do not generate embeddings for symbols")
    parser.add_argument("--base-address", help="Image base of the executable image to map for remote analysis")
    parser.add_argument("-A", action='store_true', help="Upload and Analyse a new binary")
    parser.add_argument("-u", "--upload", action='store_true', help="Upload a new binary to remote server")
    parser.add_argument("--duplicate", default=False, action='store_true', help="Duplicate an existing binary")
    parser.add_argument("-n", "--ann", action='store_true',
                        help="Fetch Approximate Nearest Neighbours (ANNs) for embedding")
    parser.add_argument("-e", "--embedding", help="Path of JSON file containing a BinNet embedding")
    parser.add_argument("--nns", default="5", help="Number of approximate nearest neighbors to fetch")
    parser.add_argument("--collections", default=None,
                        help="Comma Seperated Value of collections to search from e.g. libxml2,libpcap. Used to select RevEng.AI collections for filtering search results")
    parser.add_argument("--found-in", help="ANN flag to limit to embeddings returned to those found in specific binary")
    parser.add_argument("--from-file",
                        help="ANN flag to limit to embeddings returned to those found in JSON embeddings file")
    parser.add_argument("-c", "--cves", action="store_true", help="Check for CVEs found inside binary")
    # parser.add_argument("-C", "--sca", action="store_true",
    #                     help="Perform Software Composition Anaysis to identify common libraries embedded in binary")
    parser.add_argument("--sbom", action="store_true", help="Generate SBOM for binary")
    parser.add_argument("-m", "--model", default=None, help="AI model used to generate embeddings")
    parser.add_argument("-x", "--extract", action='store_true', help="Fetch embeddings for binary")
    parser.add_argument("--start-vaddr", help="Start virtual address of the function to extract embeddings")
    parser.add_argument("--symbol", help="Name of the symbol to extract embeddings")
    parser.add_argument("-s", "--signature", action='store_true', help="Generate a RevEng.AI binary signature")
    parser.add_argument("-S", "--similarity", action='store_true',
                        help="Compute similarity from a list of binaries. Option can be used with --from-file or -t flag with CSV of file paths. All binaries must be analysed prior to being used.")
    parser.add_argument("-t", "--to", help="CSV list of executables to compute binary similarity against")
    parser.add_argument("-M", "--match", action='store_true',
                        help="Match functions in binary file. Can be used with --confidence, --deviation, --from-file, --found-in.")
    parser.add_argument("--confidence", default="high", help="Confidence threshold used to match symbols.")
    parser.add_argument("--deviation", default=0.2,
                        help="Deviation in prediction confidence between outlier and next highest symbol. Use if confident symbol is present in binary but not matching.")
    parser.add_argument("-l", "--logs", action='store_true', help="Fetch analysis log file for binary")
    parser.add_argument("-d", "--delete", action='store_true', help="Delete all metadata associated with binary")
    parser.add_argument("-k", "--apikey", help="RevEng.AI API key")
    parser.add_argument("-h", "--host", help="Analysis Host (https://api.reveng.ai)")
    parser.add_argument("-v", "--version", action="store_true", help="Display version information")
    parser.add_argument("--help", action="help", default=argparse.SUPPRESS,
                        help=argparse._('Show this help message and exit'))
    parser.add_argument("--isa", default=None, help="Override executable ISA. Valid values are x86, x86_64, ARMv7")
    parser.add_argument("--exec-format", default=None,
                        help="Override executable format. Valid values are pe, elf, macho, raw")
    parser.add_argument("--platform", default=None,
                        help="Override OS platform. Valid values are Windows, Linux, OSX, OpenBSD")
    parser.add_argument("--dynamic-execution", default=False, action='store_true',
                        help="Enable dynamic execution in sandbox during analysis. Analysis will include any auto unpacked malware samples")
    parser.add_argument("--cmd-line-args", default="",
                        help="Command line arguments to pass when running binary sample in the sandbox. Only used when run with --dynamic-execution")
    parser.add_argument("--scope", default="private", choices=["public", "private"],
                        help="Override analysis visibility (scope). Valid values are 'public' or 'private'[DEFAULT]")
    parser.add_argument("--tags", default=None, type=str,
                        help="Assign tags to an analysis. Valid responses are tag1,tag2,tag3..")
    parser.add_argument("--priority", default=0, type=int, help="Add priority to processing queue.")
    parser.add_argument("--verbose", default=False, action='store_true', help="Set verbose output.")
    args = parser.parse_args()

    # set re_conf args
    for arg in ('apikey', 'host', 'model'):
        if getattr(args, arg):
            api.re_conf[arg] = getattr(args, arg)

    logging.basicConfig(level=logging.DEBUG if args.verbose else logging.INFO)

    # validate length of string tags
    if args.tags:
        # don't add non-content as tags
        if len(args.tags.strip()) == 0:
            args.tags = None

        else:
            # convert to list
            args.tags = args.tags.split(',')

    # display version and exit
    if args.version:
        version()
        exit(0)

    exec_fmt = None
    exec_isa = None
    base_address = 0
    if args.base_address:
        if args.base_address.upper()[:2] == "0X":
            base_address = int(args.base_address, 16)
        else:
            base_address = int(args.base_address)

    collections = None
    if args.collections:
        collections = parse_collections(args.collections)

    # auto analysis, uploads and starts analysis
    if args.A:
        args.upload = True
        args.analyse = True

    if args.dir:
        if not os.path.isdir(args.dir):
            rerr.print(f'Error, {args.dir} is not a valid directory path')
            exit(-1)

        files = iglob(os.path.abspath(args.dir) + '/**/*', recursive=True)
        ## perform operation on all files inside directory
        for file in track(files, description='Files in directory'):
            if not os.path.isfile(file):
                rerr.print(f'[blue]Skipping non-file[/blue] {file}')
                continue

            # upload binary
            if args.upload:
                api.RE_upload(file)

            if args.analyse:
                try:
                    fpath, exec_fmt, exec_isa = verify_binary(file)
                    rout.print(f'Found {fpath}:{exec_fmt}-{exec_isa}')
                    rout.print(f'[green bold]Analysing[/green bold] {file}')
                    api.RE_analyse(file, model_name=args.model, isa_options=args.isa, platform_options=args.platform,
                                   dynamic_execution=args.dynamic_execution, command_line_args=args.cmd_line_args,
                                   file_options=args.exec_format, scope=args.scope.upper(), tags=args.tags,
                                   priority=args.priority, duplicate=args.duplicate)
                except Exception as e:
                    rerr.print(f"[red bold][!] Error, binary exec type could not be verified[/red bold] {file}")

            if args.delete:
                try:
                    rout.print(f'[green bold]Deleting analyses for[/green bold] {file}')
                    api.RE_delete(file)
                except Exception as e:
                    rerr.print(f"[red bold][!] Error, could not delete analysis for [/red bold] {file}")
                    rerr.print(f"[yellow] {e} [/yellow]")
            if not (args.upload or args.analyse or args.delete):
                rerr.print(f'Error, -D only supports upload, analyse, or delete')
                exit(-1)

        exit(0)

    if args.A or args.analyse or args.extract or args.logs or args.delete or args.signature or args.similarity or args.upload or args.match or args.sbom:
        # verify binary is a file
        try:
            fpath, exec_fmt, exec_isa = verify_binary(args.binary)
            # keep stdout to data only
            rerr.print(f'Found {fpath}:{exec_fmt}-{exec_isa}')
            args.binary = fpath
        except Exception as e:
            rerr.print(f"[bold red]{str(e)}[/bold red]")
            rerr.print("[bold red][!] Error, please supply a valid binary file using '-b'.[/bold red]")
            # parser.print_help()
            exit(-1)

    if args.upload:

        api.RE_upload(args.binary)

        if not args.analyse:
            exit(0)
        # upload binary first, them carry out actions

    if args.analyse:
        api.RE_analyse(args.binary, model_name=args.model, isa_options=args.isa, platform_options=args.platform,
                       dynamic_execution=args.dynamic_execution, command_line_args=args.cmd_line_args,
                       file_options=args.exec_format, scope=args.scope.upper(), tags=args.tags, priority=args.priority,
                       duplicate=args.duplicate)

    elif args.extract:
        embeddings = api.RE_embeddings(args.binary).json()
        print_json(data=embeddings)

    elif args.signature and not args.ann:
        # Arithetic mean of symbol embeddings
        b_embed = api.RE_signature(args.binary).json()
        print_json(data=b_embed)

    elif args.similarity:
        # compute binary similarity from list of executables
        if args.from_file:
            binaries = list(map(lambda x: x.strip(), open(args.from_file, 'r').readlines()))
        else:
            if not args.to:
                print(f"Error, please specify --from-file or --to to compute binary similarity against")
                exit(-1)
            binaries = args.to.split(",")

        # verify all binaries are valid files
        for b in binaries:
            verify_binary(b)

        binary_similarity(args.binary, binaries, args.model)

    elif args.ann:
        # parse embedding json file
        if args.embedding:
            if not isfile(args.embedding):
                print("[!] Error, please supply a valid embedding JSON file using '-e'")
                parser.print_help()
                exit(-1)

            embedding = json.loads(open(args.embedding, 'r').read())
        elif (args.symbol or args.start_vaddr) and args.binary:
            if args.start_vaddr:
                if args.start_vaddr.upper()[:2] == "0X":
                    vaddr = int(args.start_vaddr, 16) + base_address
                else:
                    vaddr = int(args.start_vaddr) + base_address

                print(
                    f"[+] Using symbol starting at vaddr {hex(vaddr)} from {args.binary} (image_base:{hex(base_address)})")
                embeddings = api.RE_embeddings(args.binary).json()
                matches = list(filter(lambda x: x['vaddr'] == vaddr, embeddings))
                if len(matches) == 0:
                    print(f"[!] Error, could not find symbol at {hex(vaddr)} in {args.binary}")
                    exit(-1)
                embedding = matches[0]['embedding']
            else:
                print(f"[+] Using symbol {args.symbol} from {args.binary}")

                embeddings = api.RE_embeddings(args.binary).json()
                matches = list(filter(lambda x: x['name'] == args.symbol, embeddings))
                if len(matches) == 0:
                    print(f"[!] Error, could not find symbol at {args.symbol} in {args.binary}")
                    exit(-1)
                embedding = matches[0]['embedding']
        elif args.binary and args.signature:
            print(f"[+] Searching ANN for binary embeddings {args.binary}")
            b_suggestions = api.RE_nearest_binaries(api.RE_signature(args.binary).json(), args.model, args.nns,
                                                    collections, ignore_hashes=[api.re_binary_id(args.binary)])
            print_json(data=b_suggestions)
            exit(0)
        else:
            rerr.print("[bold red][!] Error, please supply a valid embedding JSON file using '-e', or select a function"
                       " using --start-vaddr or --symbol (NB: -b flag is needed for both of these options).[/bold red]")
            # parser.print_help()
            exit(-1)

        if args.found_in:
            if not os.path.isfile(args.found_in):
                print("[!] Error, --found-in flag requires a path to a binary to search from")
                exit(-1)
            print(f"[+] Searching for symbols similar to embedding in binary {args.found_in}")
            embeddings = api.RE_embeddings(args.found_in).json()
            res = api.RE_compute_distance(embedding, embeddings, int(args.nns))
            print_json(data=res)
        elif args.from_file:
            if not os.path.isfile(args.from_file):
                print("[!] Error, --from-file flag requires a path to a JSON embeddings file")
                exit(-1)
            print(f"[+] Searching for symbols similar to embedding in binary {args.from_file}")
            res = api.RE_compute_distance(embedding, json.load(open(args.from_file, "r")), int(args.nns))
            print_json(data=res)
        else:
            print(f"[+] Searching for similar symbols to embedding in "
                  f"{'all' if not args.collections else args.collections} collections.")
            f_suggestions = api.RE_nearest_symbols(embedding["embedding"], args.model, int(args.nns),
                                                   collections=collections).json()
            print_json(data=f_suggestions)

    elif args.match:
        # parse confidences
        confidence = 0.90
        if args.confidence:
            confidences = {
                'high': 0.95,
                'medium': 0.9,
                'low': 0.7,
                'partial': 0.5,
                'all': 0.0
            }
            if args.confidence in confidences.keys():
                confidence = confidences[args.confidence]
            else:
                confidence = float(args.confidence)

        if args.from_file:
            embeddings = json.load(open(args.from_file, 'r'))
        elif args.found_in:
            if not os.path.isfile(args.found_in):
                print("[!] Error, --found-in flag requires a path to a binary to search from")
                exit(-1)
            print(f"[+] Matching symbols between {args.binary} and {args.found_in}")
            embeddings = api.RE_embeddings(args.found_in).json()
        else:
            # print("No --from-file or --found-in, matching from global symbol database (unstrip) not currently")
            match_for_each(args.binary, args.model, confidence, collections)
            exit(-1)

        match(args.binary, args.model, embeddings, confidence=confidence, deviation=float(args.deviation))

    # elif args.sca:
    #     api.RE_sca(args.binary)

    elif args.logs:
        api.RE_logs(args.binary)

    elif args.delete:
        api.RE_delete(args.binary)

    elif args.sbom:
        api.RE_SBOM(args.binary)

    elif args.cves:
        api.RE_cves(args.binary)
    else:
        print("[!] Error, please supply an action command")
        parser.print_help()


if __name__ == '__main__':
    main()
